@@ -24,10 +24,10 @@
 // contributors may be used to endorse or promote products derived from
 // this software without specific prior written permission.
 //
-// THIS SOFTWARE IS PROVIDED BY SANDIA CORPORATION "AS IS" AND ANY
+// THIS SOFTWARE IS PROVIDED BY NTESS "AS IS" AND ANY
 // EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 // IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
-// PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL SANDIA CORPORATION OR THE
+// PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NTESS OR THE
 // CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 // EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 // PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
@@ -50,12 +50,13 @@
 #include <utility>
 #include <iostream>
 #include <sstream>
+#include <thread>
 
 #include <Kokkos_Core.hpp>
 
 #include <impl/Kokkos_Error.hpp>
 #include <impl/Kokkos_CPUDiscovery.hpp>
-#include <impl/Kokkos_Profiling_Interface.hpp>
+#include <impl/Kokkos_Tools.hpp>
 
 //----------------------------------------------------------------------------
 //----------------------------------------------------------------------------
@@ -65,8 +66,8 @@ namespace Impl {
 namespace {
 
 ThreadsExec s_threads_process;
-ThreadsExec *s_threads_exec[ThreadsExec::MAX_THREAD_COUNT] = {0};
-pthread_t s_threads_pid[ThreadsExec::MAX_THREAD_COUNT]     = {0};
+ThreadsExec *s_threads_exec[ThreadsExec::MAX_THREAD_COUNT] = {nullptr};
+std::thread::id s_threads_pid[ThreadsExec::MAX_THREAD_COUNT];
 std::pair<unsigned, unsigned> s_threads_coord[ThreadsExec::MAX_THREAD_COUNT];
 
 int s_thread_pool_size[3] = {0, 0, 0};
@@ -75,11 +76,9 @@ unsigned s_current_reduce_size = 0;
 unsigned s_current_shared_size = 0;
 
 void (*volatile s_current_function)(ThreadsExec &, const void *);
-const void *volatile s_current_function_arg = 0;
+const void *volatile s_current_function_arg = nullptr;
 
 struct Sentinel {
-  Sentinel() {}
-
   ~Sentinel() {
     if (s_thread_pool_size[0] || s_thread_pool_size[1] ||
         s_thread_pool_size[2] || s_current_reduce_size ||
@@ -113,7 +112,7 @@ namespace Impl {
 
 void execute_function_noop(ThreadsExec &, const void *) {}
 
-void ThreadsExec::driver(void) {
+void ThreadsExec::driver() {
   SharedAllocationRecord<void, void>::tracking_enable();
 
   ThreadsExec this_thread;
@@ -129,8 +128,8 @@ void ThreadsExec::driver(void) {
 }
 
 ThreadsExec::ThreadsExec()
-    : m_pool_base(0),
-      m_scratch(0),
+    : m_pool_base(nullptr),
+      m_scratch(nullptr),
       m_scratch_reduce_end(0),
       m_scratch_thread_end(0),
       m_numa_rank(0),
@@ -142,7 +141,7 @@ ThreadsExec::ThreadsExec()
   if (&s_threads_process != this) {
     // A spawned thread
 
-    ThreadsExec *const nil = 0;
+    ThreadsExec *const nil = nullptr;
 
     // Which entry in 's_threads_exec', possibly determined from hwloc binding
     const int entry =
@@ -166,7 +165,7 @@ ThreadsExec::ThreadsExec()
       m_pool_fan_size  = fan_size(m_pool_rank, m_pool_size);
       m_pool_state     = ThreadsExec::Active;
 
-      s_threads_pid[m_pool_rank] = pthread_self();
+      s_threads_pid[m_pool_rank] = std::this_thread::get_id();
 
       // Inform spawning process that the threads_exec entry has been set.
       s_threads_process.m_pool_state = ThreadsExec::Active;
@@ -180,24 +179,24 @@ ThreadsExec::ThreadsExec()
     m_pool_size  = 1;
     m_pool_state = ThreadsExec::Inactive;
 
-    s_threads_pid[m_pool_rank] = pthread_self();
+    s_threads_pid[m_pool_rank] = std::this_thread::get_id();
   }
 }
 
 ThreadsExec::~ThreadsExec() {
   const unsigned entry = m_pool_size - (m_pool_rank + 1);
 
-  typedef Kokkos::Impl::SharedAllocationRecord<Kokkos::HostSpace, void> Record;
+  using Record = Kokkos::Impl::SharedAllocationRecord<Kokkos::HostSpace, void>;
 
   if (m_scratch) {
     Record *const r = Record::get_record(m_scratch);
 
-    m_scratch = 0;
+    m_scratch = nullptr;
 
     Record::decrement(r);
   }
 
-  m_pool_base          = 0;
+  m_pool_base          = nullptr;
   m_scratch_reduce_end = 0;
   m_scratch_thread_end = 0;
   m_numa_rank          = 0;
@@ -209,7 +208,7 @@ ThreadsExec::~ThreadsExec() {
   m_pool_state = ThreadsExec::Terminating;
 
   if (&s_threads_process != this && entry < MAX_THREAD_COUNT) {
-    ThreadsExec *const nil = 0;
+    ThreadsExec *const nil = nullptr;
 
     atomic_compare_exchange(s_threads_exec + entry, this, nil);
 
@@ -223,13 +222,13 @@ ThreadsExec *ThreadsExec::get_thread(const int init_thread_rank) {
   ThreadsExec *const th =
       init_thread_rank < s_thread_pool_size[0]
           ? s_threads_exec[s_thread_pool_size[0] - (init_thread_rank + 1)]
-          : 0;
+          : nullptr;
 
-  if (0 == th || th->m_pool_rank != init_thread_rank) {
+  if (nullptr == th || th->m_pool_rank != init_thread_rank) {
     std::ostringstream msg;
     msg << "Kokkos::Impl::ThreadsExec::get_thread ERROR : "
         << "thread " << init_thread_rank << " of " << s_thread_pool_size[0];
-    if (0 == th) {
+    if (nullptr == th) {
       msg << " does not exist";
     } else {
       msg << " has wrong thread_rank " << th->m_pool_rank;
@@ -290,21 +289,46 @@ int ThreadsExec::in_parallel() {
   return s_current_function && (&s_threads_process != s_current_function_arg) &&
          (s_threads_process.m_pool_base || !is_process());
 }
+void ThreadsExec::fence() { internal_fence(Impl::fence_is_static::yes); }
+void ThreadsExec::fence(const std::string &name) {
+  internal_fence(name, Impl::fence_is_static::yes);
+}
+
+void ThreadsExec::internal_fence(Impl::fence_is_static is_static) {
+  internal_fence((is_static == Impl::fence_is_static::no)
+                     ? "Kokkos::ThreadsExec::fence: Unnamed Instance Fence"
+                     : "Kokkos::ThreadsExec::fence: Unnamed Global Fence",
+                 is_static);
+}
 
 // Wait for root thread to become inactive
-void ThreadsExec::fence() {
-  if (s_thread_pool_size[0]) {
-    // Wait for the root thread to complete:
-    Impl::spinwait_while_equal<int>(s_threads_exec[0]->m_pool_state,
-                                    ThreadsExec::Active);
+void ThreadsExec::internal_fence(const std::string &name,
+                                 Impl::fence_is_static is_static) {
+  const auto &fence_lam = [&]() {
+    if (s_thread_pool_size[0]) {
+      // Wait for the root thread to complete:
+      Impl::spinwait_while_equal<int>(s_threads_exec[0]->m_pool_state,
+                                      ThreadsExec::Active);
+    }
+
+    s_current_function     = nullptr;
+    s_current_function_arg = nullptr;
+
+    // Make sure function and arguments are cleared before
+    // potentially re-activating threads with a subsequent launch.
+    memory_fence();
+  };
+  if (is_static == Impl::fence_is_static::yes) {
+    Kokkos::Tools::Experimental::Impl::profile_fence_event<Kokkos::Threads>(
+        name,
+        Kokkos::Tools::Experimental::SpecialSynchronizationCases::
+            GlobalDeviceSynchronization,
+        fence_lam);
+  } else {
+    Kokkos::Tools::Experimental::Impl::profile_fence_event<Kokkos::Threads>(
+        name, Kokkos::Tools::Experimental::Impl::DirectFenceIDHandle{1},
+        fence_lam);
   }
-
-  s_current_function     = 0;
-  s_current_function_arg = 0;
-
-  // Make sure function and arguments are cleared before
-  // potentially re-activating threads with a subsequent launch.
-  memory_fence();
 }
 
 /** \brief  Begin execution of the asynchronous functor */
@@ -364,7 +388,7 @@ bool ThreadsExec::wake() {
   ThreadsExec::global_unlock();
 
   if (s_threads_process.m_pool_base) {
-    execute_sleep(s_threads_process, 0);
+    execute_sleep(s_threads_process, nullptr);
     s_threads_process.m_pool_state = ThreadsExec::Inactive;
   }
 
@@ -375,15 +399,30 @@ bool ThreadsExec::wake() {
 
 //----------------------------------------------------------------------------
 
-void ThreadsExec::execute_serial(void (*func)(ThreadsExec &, const void *)) {
-  s_current_function     = func;
+void ThreadsExec::execute_resize_scratch_in_serial() {
+  const unsigned begin = s_threads_process.m_pool_base ? 1 : 0;
+
+  auto deallocate_scratch_memory = [](ThreadsExec &exec) {
+    if (exec.m_scratch) {
+      using Record =
+          Kokkos::Impl::SharedAllocationRecord<Kokkos::HostSpace, void>;
+      Record *const r = Record::get_record(exec.m_scratch);
+      exec.m_scratch  = nullptr;
+      Record::decrement(r);
+    }
+  };
+  if (s_threads_process.m_pool_base) {
+    for (unsigned i = s_thread_pool_size[0]; begin < i;) {
+      deallocate_scratch_memory(*s_threads_exec[--i]);
+    }
+  }
+
+  s_current_function     = &first_touch_allocate_thread_private_scratch;
   s_current_function_arg = &s_threads_process;
 
   // Make sure function and arguments are written before activating threads.
   memory_fence();
 
-  const unsigned begin = s_threads_process.m_pool_base ? 1 : 0;
-
   for (unsigned i = s_thread_pool_size[0]; begin < i;) {
     ThreadsExec &th = *s_threads_exec[--i];
 
@@ -393,13 +432,14 @@ void ThreadsExec::execute_serial(void (*func)(ThreadsExec &, const void *)) {
   }
 
   if (s_threads_process.m_pool_base) {
+    deallocate_scratch_memory(s_threads_process);
     s_threads_process.m_pool_state = ThreadsExec::Active;
-    (*func)(s_threads_process, 0);
+    first_touch_allocate_thread_private_scratch(s_threads_process, nullptr);
     s_threads_process.m_pool_state = ThreadsExec::Inactive;
   }
 
-  s_current_function_arg = 0;
-  s_current_function     = 0;
+  s_current_function_arg = nullptr;
+  s_current_function     = nullptr;
 
   // Make sure function and arguments are cleared before proceeding.
   memory_fence();
@@ -411,25 +451,18 @@ void *ThreadsExec::root_reduce_scratch() {
   return s_threads_process.reduce_memory();
 }
 
-void ThreadsExec::execute_resize_scratch(ThreadsExec &exec, const void *) {
-  typedef Kokkos::Impl::SharedAllocationRecord<Kokkos::HostSpace, void> Record;
-
-  if (exec.m_scratch) {
-    Record *const r = Record::get_record(exec.m_scratch);
-
-    exec.m_scratch = 0;
-
-    Record::decrement(r);
-  }
-
+void ThreadsExec::first_touch_allocate_thread_private_scratch(ThreadsExec &exec,
+                                                              const void *) {
   exec.m_scratch_reduce_end = s_threads_process.m_scratch_reduce_end;
   exec.m_scratch_thread_end = s_threads_process.m_scratch_thread_end;
 
   if (s_threads_process.m_scratch_thread_end) {
     // Allocate tracked memory:
     {
+      using Record =
+          Kokkos::Impl::SharedAllocationRecord<Kokkos::HostSpace, void>;
       Record *const r =
-          Record::allocate(Kokkos::HostSpace(), "thread_scratch",
+          Record::allocate(Kokkos::HostSpace(), "Kokkos::thread_scratch",
                            s_threads_process.m_scratch_thread_end);
 
       Record::increment(r);
@@ -469,7 +502,7 @@ void *ThreadsExec::resize_scratch(size_t reduce_size, size_t thread_size) {
     s_threads_process.m_scratch_reduce_end = reduce_size;
     s_threads_process.m_scratch_thread_end = reduce_size + thread_size;
 
-    execute_serial(&execute_resize_scratch);
+    execute_resize_scratch_in_serial();
 
     s_threads_process.m_scratch = s_threads_exec[0]->m_scratch;
   }
@@ -508,7 +541,7 @@ void ThreadsExec::print_configuration(std::ostream &s, const bool detail) {
     s << " threads[" << s_thread_pool_size[0] << "]"
       << " threads_per_numa[" << s_thread_pool_size[1] << "]"
       << " threads_per_core[" << s_thread_pool_size[2] << "]";
-    if (0 == s_threads_process.m_pool_base) {
+    if (nullptr == s_threads_process.m_pool_base) {
       s << " Asynchronous";
     }
     s << " ReduceScratch[" << s_current_reduce_size << "]"
@@ -547,19 +580,20 @@ void ThreadsExec::print_configuration(std::ostream &s, const bool detail) {
 
 //----------------------------------------------------------------------------
 
-int ThreadsExec::is_initialized() { return 0 != s_threads_exec[0]; }
+int ThreadsExec::is_initialized() { return nullptr != s_threads_exec[0]; }
 
 void ThreadsExec::initialize(unsigned thread_count, unsigned use_numa_count,
                              unsigned use_cores_per_numa,
                              bool allow_asynchronous_threadpool) {
-  static const Sentinel sentinel;
+  // need to provide an initializer for Intel compilers
+  static const Sentinel sentinel = {};
 
   const bool is_initialized = 0 != s_thread_pool_size[0];
 
   unsigned thread_spawn_failed = 0;
 
   for (int i = 0; i < ThreadsExec::MAX_THREAD_COUNT; i++)
-    s_threads_exec[i] = NULL;
+    s_threads_exec[i] = nullptr;
 
   if (!is_initialized) {
     // If thread_count, use_numa_count, or use_cores_per_numa are zero
@@ -611,11 +645,10 @@ void ThreadsExec::initialize(unsigned thread_count, unsigned use_numa_count,
 
       // Spawn thread executing the 'driver()' function.
       // Wait until spawned thread has attempted to initialize.
-      // If spawning and initialization is successfull then
+      // If spawning and initialization is successful then
       // an entry in 's_threads_exec' will be assigned.
-      if (ThreadsExec::spawn()) {
-        wait_yield(s_threads_process.m_pool_state, ThreadsExec::Inactive);
-      }
+      ThreadsExec::spawn();
+      wait_yield(s_threads_process.m_pool_state, ThreadsExec::Inactive);
       if (s_threads_process.m_pool_state == ThreadsExec::Terminating) break;
     }
 
@@ -631,15 +664,15 @@ void ThreadsExec::initialize(unsigned thread_count, unsigned use_numa_count,
       }
     }
 
-    s_current_function             = 0;
-    s_current_function_arg         = 0;
+    s_current_function             = nullptr;
+    s_current_function_arg         = nullptr;
     s_threads_process.m_pool_state = ThreadsExec::Inactive;
 
     memory_fence();
 
     if (!thread_spawn_failed) {
       // Bind process to the core on which it was located before spawning
-      // occured
+      // occurred
       if (hwloc_can_bind) {
         Kokkos::hwloc::bind_this_thread(proc_coord);
       }
@@ -657,9 +690,10 @@ void ThreadsExec::initialize(unsigned thread_count, unsigned use_numa_count,
         s_threads_process.m_pool_size     = thread_count;
         s_threads_process.m_pool_fan_size = fan_size(
             s_threads_process.m_pool_rank, s_threads_process.m_pool_size);
-        s_threads_pid[s_threads_process.m_pool_rank] = pthread_self();
+        s_threads_pid[s_threads_process.m_pool_rank] =
+            std::this_thread::get_id();
       } else {
-        s_threads_process.m_pool_base     = 0;
+        s_threads_process.m_pool_base     = nullptr;
         s_threads_process.m_pool_rank     = 0;
         s_threads_process.m_pool_size     = 0;
         s_threads_process.m_pool_fan_size = 0;
@@ -709,10 +743,6 @@ void ThreadsExec::initialize(unsigned thread_count, unsigned use_numa_count,
   Impl::init_lock_array_host_space();
 
   Impl::SharedAllocationRecord<void, void>::tracking_enable();
-
-#if defined(KOKKOS_ENABLE_DEPRECATED_CODE) && defined(KOKKOS_ENABLE_PROFILING)
-  Kokkos::Profiling::initialize();
-#endif
 }
 
 //----------------------------------------------------------------------------
@@ -735,12 +765,12 @@ void ThreadsExec::finalize() {
       s_threads_process.m_pool_state = ThreadsExec::Inactive;
     }
 
-    s_threads_pid[i] = 0;
+    s_threads_pid[i] = std::thread::id();
   }
 
   if (s_threads_process.m_pool_base) {
     (&s_threads_process)->~ThreadsExec();
-    s_threads_exec[0] = 0;
+    s_threads_exec[0] = nullptr;
   }
 
   if (Kokkos::hwloc::can_bind_threads()) {
@@ -754,15 +784,13 @@ void ThreadsExec::finalize() {
   // Reset master thread to run solo.
   s_threads_process.m_numa_rank      = 0;
   s_threads_process.m_numa_core_rank = 0;
-  s_threads_process.m_pool_base      = 0;
+  s_threads_process.m_pool_base      = nullptr;
   s_threads_process.m_pool_rank      = 0;
   s_threads_process.m_pool_size      = 1;
   s_threads_process.m_pool_fan_size  = 0;
   s_threads_process.m_pool_state     = ThreadsExec::Inactive;
 
-#if defined(KOKKOS_ENABLE_PROFILING)
   Kokkos::Profiling::finalize();
-#endif
 }
 
 //----------------------------------------------------------------------------
@@ -775,55 +803,99 @@ void ThreadsExec::finalize() {
 
 namespace Kokkos {
 
-int Threads::concurrency() {
-#ifdef KOKKOS_ENABLE_DEPRECATED_CODE
-  return thread_pool_size(0);
-#else
-  return impl_thread_pool_size(0);
-#endif
+int Threads::concurrency() { return impl_thread_pool_size(0); }
+void Threads::fence() const {
+  Impl::ThreadsExec::internal_fence(Impl::fence_is_static::no);
+}
+void Threads::fence(const std::string &name) const {
+  Impl::ThreadsExec::internal_fence(name, Impl::fence_is_static::no);
 }
-#ifndef KOKKOS_ENABLE_DEPRECATED_CODE
-void Threads::fence() const { Impl::ThreadsExec::fence(); }
-#endif
 
-#ifdef KOKKOS_ENABLE_DEPRECATED_CODE
-Threads &Threads::instance(int)
-#else
-Threads &Threads::impl_instance(int)
-#endif
-{
+Threads &Threads::impl_instance(int) {
   static Threads t;
   return t;
 }
 
-#ifdef KOKKOS_ENABLE_DEPRECATED_CODE
-int Threads::thread_pool_size(int depth)
-#else
-int Threads::impl_thread_pool_size(int depth)
-#endif
-{
-  return Impl::s_thread_pool_size[depth];
-}
-
-#if defined(KOKKOS_ACTIVE_EXECUTION_MEMORY_SPACE_HOST)
-#ifdef KOKKOS_ENABLE_DEPRECATED_CODE
-int Threads::thread_pool_rank()
-#else
-int Threads::impl_thread_pool_rank()
-#endif
-{
-  const pthread_t pid = pthread_self();
-  int i               = 0;
+int Threads::impl_thread_pool_rank_host() {
+  const std::thread::id pid = std::this_thread::get_id();
+  int i                     = 0;
   while ((i < Impl::s_thread_pool_size[0]) && (pid != Impl::s_threads_pid[i])) {
     ++i;
   }
   return i;
 }
-#endif
+
+int Threads::impl_thread_pool_size(int depth) {
+  return Impl::s_thread_pool_size[depth];
+}
 
 const char *Threads::name() { return "Threads"; }
-} /* namespace Kokkos */
 
+namespace Impl {
+
+int g_threads_space_factory_initialized =
+    initialize_space_factory<ThreadsSpaceInitializer>("050_Threads");
+
+void ThreadsSpaceInitializer::initialize(const InitArguments &args) {
+  const int num_threads = args.num_threads;
+  const int use_numa    = args.num_numa;
+  if (std::is_same<Kokkos::Threads, Kokkos::DefaultExecutionSpace>::value ||
+      std::is_same<Kokkos::Threads,
+                   Kokkos::HostSpace::execution_space>::value) {
+    if (num_threads > 0) {
+      if (use_numa > 0) {
+        Kokkos::Threads::impl_initialize(num_threads, use_numa);
+      } else {
+        Kokkos::Threads::impl_initialize(num_threads);
+      }
+    } else {
+      Kokkos::Threads::impl_initialize();
+    }
+    // std::cout << "Kokkos::initialize() fyi: CppThread enabled and
+    // initialized"
+    // << std::endl ;
+  } else {
+    // std::cout << "Kokkos::initialize() fyi: CppThread enabled but not
+    // initialized" << std::endl ;
+  }
+}
+
+void ThreadsSpaceInitializer::finalize(const bool all_spaces) {
+  if (std::is_same<Kokkos::Threads, Kokkos::DefaultExecutionSpace>::value ||
+      std::is_same<Kokkos::Threads,
+                   Kokkos::HostSpace::execution_space>::value ||
+      all_spaces) {
+    if (Kokkos::Threads::impl_is_initialized())
+      Kokkos::Threads::impl_finalize();
+  }
+}
+
+void ThreadsSpaceInitializer::fence() { Kokkos::Threads::impl_static_fence(); }
+void ThreadsSpaceInitializer::fence(const std::string &name) {
+  Kokkos::Threads::impl_static_fence(name);
+}
+
+void ThreadsSpaceInitializer::print_configuration(std::ostream &msg,
+                                                  const bool detail) {
+  msg << "Host Parallel Execution Space:" << std::endl;
+  msg << "  KOKKOS_ENABLE_THREADS: ";
+  msg << "yes" << std::endl;
+
+  msg << "\nThreads Runtime Configuration:" << std::endl;
+  Kokkos::Threads::print_configuration(msg, detail);
+}
+
+}  // namespace Impl
+
+#ifdef KOKKOS_ENABLE_CXX14
+namespace Tools {
+namespace Experimental {
+constexpr DeviceType DeviceTypeTraits<Threads>::id;
+}
+}  // namespace Tools
+#endif
+
+} /* namespace Kokkos */
 //----------------------------------------------------------------------------
 //----------------------------------------------------------------------------
 #else
